/***************************************************************************
 *   Copyright (C) 2005-2019 by the Quassel Project                        *
 *   devel@quassel-irc.org                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) version 3.                                           *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.         *
 ***************************************************************************/

#pragma once

#include "syncableobject.h"

/*
 * gather various information about the core.
 */

class CoreInfo : public SyncableObject
{
    Q_OBJECT
    SYNCABLE_OBJECT

    Q_PROPERTY(QVariantMap coreData READ coreData WRITE setCoreData)

public:
    explicit CoreInfo(QObject *parent = nullptr);
    inline QVariant &at(const QString &key) { return _coreData[key]; }

    void setConnectedClientData(int, QVariantList);

    /**
     * Reset the core info state, clearing anything saved
     */
    void reset();

signals:
    /**
     * Signals that core information has changed
     */
    void coreDataChanged(QVariantMap);

public slots:
    QVariantMap coreData() const;
    void setCoreData(const QVariantMap &);

private:
    QVariantMap _coreData;
};
